' ******************************************************************
' Random number generators
' ******************************************************************

' ------------------------------------------------------------------
' Global constant and variables
' ------------------------------------------------------------------

ENUM RNG_Type
  RNG_MWC      ' Multiply-With-Carry
  RNG_MT       ' Mersenne Twister
  RNG_UVAG     ' Universal Virtual Array Generator
END ENUM

CONST Z  = 1.0# / 4294967296.0#        ' 1 / 2^32
CONST Z1 = 1.0# / 4294967295.0#        ' 1 / (2^32 - 1)
CONST Z2 = 1.0# / 9007199254740992.0#

DIM SHARED AS RNG_Type gRNG = RNG_MT

' ------------------------------------------------------------------
' External functions
' ------------------------------------------------------------------

DECLARE SUB InitMWC(Seed AS UINTEGER)
DECLARE SUB InitMT(Seed AS UINTEGER)
DECLARE SUB InitUVAG(Seed AS UINTEGER)

DECLARE FUNCTION IRanMWC() AS UINTEGER
DECLARE FUNCTION IRanMT() AS UINTEGER
DECLARE FUNCTION IRanUVAG() AS UINTEGER

' ******************************************************************

SUB SetRNG(RNG AS RNG_Type)
' Select generator

  gRNG = RNG
END SUB

SUB InitGen(Seed AS UINTEGER)
' Initialize generator

  SELECT CASE gRNG
    CASE RNG_MWC 
      InitMWC(Seed)
    CASE RNG_MT
      InitMT(Seed)
    CASE RNG_UVAG
      InitUVAG(Seed)
  END SELECT
END SUB

FUNCTION IRanGen() AS UINTEGER
' 32-bit random integer

  SELECT CASE gRNG
    CASE RNG_MWC 
      IRanGen = IRanMWC()
    CASE RNG_MT
      IRanGen = IRanMT()
    CASE RNG_UVAG
      IRanGen = IRanUVAG()
  END SELECT
END FUNCTION

FUNCTION IRanGen31() AS UINTEGER
' 31-bit random integer

  IRanGen31 = CLNG(IRanGen() SHR 1)
END FUNCTION

FUNCTION RanGen1() AS DOUBLE
' 32-bit random real in [0,1]

  RanGen1 = IRanGen() * Z1
END FUNCTION

FUNCTION RanGen2() AS DOUBLE
' 32-bit random real in [0,1)

  RanGen2 = IRanGen() * Z
END FUNCTION

FUNCTION RanGen3() AS DOUBLE
' 32-bit random real in (0,1)

  RanGen3 = (CDBL(IRanGen()) + 0.5#) * Z
END FUNCTION

FUNCTION RanGen53() AS DOUBLE
' 53-bit random real in [0,1)

  DIM A AS UINTEGER = IRanGen() SHR 5
  DIM B AS UINTEGER = IRanGen() SHR 6

  RanGen53 = (A * 67108864.0# + B) * Z2
END FUNCTION
